_remote_command = lambda magic_word: 'remote:{}'.format(magic_word)

class Message(object):
    def __init__(self, content, additional = None):
        self.__content    = content
        self.__additional = additional

    @property
    def content(self):
        return self.__content

    @property
    def additional(self):
        return self.__additional

class RemoteSignal(object):
    PING   = _remote_command('ping')
    PAUSE  = _remote_command('pause')
    RESUME = _remote_command('resume')



import logging

_logger      = None
_log_handler = None


def log(level, *args, **kwargs):
    global _logger

    if not _logger:
        prepare_logger(logging.ERROR)

    getattr(_logger, level)(*args, **kwargs)


def prepare_logger(level = logging.ERROR, use_fancy_format = False):
    global _logger
    global _log_handler

    if not _logger:
        _logger = logging.Logger('vireo')

    if not _log_handler:
        try:
            # Python 3
            formatter = logging.Formatter(
                '{asctime} | {name} | [{levelname}] {message}'
                if use_fancy_format
                else '{asctime} | [{levelname}] {message}',
                style='{'
            )
        except TypeError:
            # Python 2
            formatter = logging.Formatter(
                '$(name)a | [%(levelname)s] %(message)s'
                if use_fancy_format
                else '[%(levelname)s] %(message)s'
            )

        _log_handler = logging.StreamHandler()
        _log_handler.setFormatter(formatter)

        _logger.addHandler(_log_handler)

    _logger.setLevel(level)
    _log_handler.setLevel(level)

    return _logger

def fill_in_the_blank(main_dict, default_dict):
    for default_key, default_value in default_dict.items():
        if default_key in main_dict:
            continue

        main_dict[default_key] = default_value

    return main_dict



import json
import threading
import time
import uuid

from pika            import BasicProperties
from pika.exceptions import ConnectionClosed, ChannelClosed

from ...helper import fill_in_the_blank, log

from .consumer  import Consumer
from .exception import NoConnectionError, SubscriptionNotAllowedError
from .helper    import active_connection, SHARED_DIRECT_EXCHANGE_NAME, SHARED_TOPIC_EXCHANGE_NAME, SHARED_SIGNAL_CONNECTION_LOSS


class Driver(object):
    """ Driver for RabbitMQ

        :param          url:               the URL to the server (``str`` for a single connection or ``list`` for rotation)
        :param list     consumer_classes:  the list of :class:`.consumer.Consumer`-based classes
        :param bool     unlimited_retries: the flag to disable limited retry count.
        :param callable on_connect:        a callback function when the message consumption begins.
        :param callable on_disconnect:     a callback function when the message consumption is interrupted due to unexpected disconnection.
        :param callable on_error:          a callback function when the message consumption is interrupted due to exception raised from the main callback function.

        :param dict default_publishing_options:                  the default options for publishing (normal)
        :param dict default_broadcasting_options:                the default options for publishing (broadcast)
        :param dict default_consuming_shared_queue_options:      the default options for consuming share queue
        :param dict default_consuming_distributed_queue_options: the default options for consuming distributed queue

        ``default_publishing_options`` and ``default_broadcasting_options`` only take ``exchange``
        to allow overriding the default exchange.

        ``default_consuming_shared_queue_options`` and ``default_consuming_distributed_queue_options``
        will have the data structure like this::

            {
                'exchange': {
                    'name': str, # It is "exchange" in pika's exchange_declare.
                    'type': str, # It is "exchange_type" in pika's exchange_declare.
                }
            }

        Here is an example for ``on_connect``.

        .. code-block:: Python

            def on_connect(consumer = None):
                ...

        Here is an example for ``on_disconnect``.

        .. code-block:: Python

            def on_disconnect(consumer = None):
                ...

        Here is an example for ``on_error``.

        .. code-block:: Python

            def on_error(exception, consumer = None):
                ...
    """
    def __init__(self, url, consumer_classes = None, unlimited_retries = False, on_connect = None,
                 on_disconnect = None, on_error = None, default_publishing_options : dict = None,
                 default_broadcasting_options : dict = None, default_consuming_shared_queue_options : dict = None,
                 default_consuming_distributed_queue_options : dict = None, auto_acknowledge = False,
                 send_sigterm_on_disconnect = True):
        for consumer_class in consumer_classes or []:
            assert isinstance(consumer_class, Consumer), 'This ({}) needs to be a subclass of vireo.drivers.rabbitmq.Consumer.'.format(consumer_class)

        self._url              = url
        self._consumer_classes = consumer_classes or []
        self._async_listener   = None
        self._shared_stream    = []
        self._consumers        = []
        self._has_term_signal  = False
        self._active_routes    = []
        self._auto_acknowledge = auto_acknowledge

        self._send_sigterm_on_disconnect = send_sigterm_on_disconnect

        self._default_publishing_options                  = default_publishing_options                  or {}
        self._default_broadcasting_options                = default_broadcasting_options                or {}
        self._default_consuming_shared_queue_options      = default_consuming_shared_queue_options      or {}
        self._default_consuming_distributed_queue_options = default_consuming_distributed_queue_options or {}

        self._unlimited_retries = unlimited_retries
        self._on_connect        = on_connect
        self._on_disconnect     = on_disconnect
        self._on_error          = on_error

        self._requested_url_counter = 0 # used for rotation
        self._total_url_count       = 1 if isinstance(self._url, str) else len(self._url)

    @property
    def url(self):
        connection_url = self._url

        if isinstance(self._url, (tuple, list)):
            self._requested_url_counter += 1

            connection_index = self._requested_url_counter % self._total_url_count
            connection_url   = self._url[connection_index]

        return connection_url

    def set_on_connect(self, on_connect):
        self._on_connect = on_connect

    def set_on_disconnect(self, on_disconnect):
        self._on_disconnect = on_disconnect

    def set_on_error(self, on_error):
        self._on_error = on_error

    def setup_async_cleanup(self):
        """ Prepare to cleanly join all consumers asynchronously. """
        if self._async_listener and self._async_listener.is_alive():
            raise SubscriptionNotAllowedError('Unable to consume messages as this driver is currently active.')

        self._async_listener = threading.Thread(target = self.join)
        self._async_listener.start()

    def stop_consuming(self):
        """ Send the signal to stop consumption. """
        self._has_term_signal = True

    def join(self):
        """ Synchronously join all consumers."""
        try:
            while True:
                if self._has_term_signal:
                    log('warning', 'Stopping all route listeners')

                    break

                if SHARED_SIGNAL_CONNECTION_LOSS in self._shared_stream:
                    log('error', 'Unexpected connection loss detected')
                    log('warning', 'Terminating all route listeners')

                    break

                time.sleep(1)
        except KeyboardInterrupt:
            log('warning', 'SIGTERM received')
            log('debug', 'Terminating all route listeners')

        connection_losed = SHARED_SIGNAL_CONNECTION_LOSS in self._shared_stream

        for consumer in self._consumers:
            if not connection_losed:
                if not consumer.is_alive():
                    log('info', 'Route {}: Already stopped listening (not alive).'.format(consumer.route))

                    continue

                log('warning', 'Route {}: Sending the signal to stop listening.'.format(consumer.route))
                consumer.stop()

            try:
                log('debug', 'Route {}: Terminating the listener.'.format(consumer.route))
                consumer._stop()
            except AssertionError: # this is raised if the thread lock is still locked.
                log('warning', 'Route {}: Probably already stopped'.format(consumer.route))

            if not consumer.is_alive():
                log('info', 'Route {}: Termination confirmed (killed)'.format(consumer.route))

                continue

            log('debug', 'Route {}: Waiting the listener to join back to the parent thread.'.format(consumer.route))
            consumer.join()
            log('info', 'Route {}: Termination confirmed (joined).'.format(consumer.route))

        if connection_losed:
            raise NoConnectionError('Unexpectedly losed the connection during message consumption')

    def publish(self, route, message, options = None):
        """ Synchronously publish a message

            :param str route:   the route
            :param str message: the message
            :param dict options: additional options for basic_publish
        """
        default_parameters = self._generate_default_publish_options(
            self._default_publishing_options,
            SHARED_DIRECT_EXCHANGE_NAME,
            route,
            message,
        )

        options = fill_in_the_blank(options or {}, default_parameters)

        with active_connection(self.url, self._on_connect, self._on_disconnect) as channel:
            try:
                log('debug', 'Publishing: route={} message={} options={}'.format(route, message, options))
                channel.basic_publish(**options)
                log('debug', 'Published: route={} message={} options={}'.format(route, message, options))
            except ConnectionClosed:
                raise NoConnectionError('Unexpectedly losed the connection while publishing a message')

    def declare_queue_with_delegation(self, origin_queue_name, ttl, fallback_queue_name = None,
                                      common_queue_options = None, exchange_options = None):
        actual_fallback_queue_name = fallback_queue_name or '{}.delegated'.format(origin_queue_name)
        exchange_name              = 'fallback/{}/{}'.format(origin_queue_name, actual_fallback_queue_name)

        exchange_options     = fill_in_the_blank(exchange_options or {}, {'internal': True})
        common_queue_options = common_queue_options or {}

        default_fallback_queue_options = {'auto_delete': False}
        default_origin_queue_options   = {
            'auto_delete' : False,
            'arguments'   : {
                'x-dead-letter-exchange'    : exchange_name,
                'x-dead-letter-routing-key' : actual_fallback_queue_name,
                'x-message-ttl'             : ttl,
            }
        }

        fallback_queue_options = fill_in_the_blank(dict(common_queue_options), default_fallback_queue_options)
        origin_queue_options   = fill_in_the_blank(dict(common_queue_options), default_origin_queue_options)

        fill_in_the_blank(exchange_options, {'exchange': exchange_name, 'exchange_type': 'direct'})

        with active_connection(self.url, self._on_connect, self._on_disconnect) as channel:
            try:
                channel.exchange_declare(**exchange_options)

                self.declare_queue(actual_fallback_queue_name, fallback_queue_options)

                channel.queue_bind(
                    queue    = actual_fallback_queue_name,
                    exchange = exchange_name,
                )

                self.declare_queue(origin_queue_name, origin_queue_options)
            except ConnectionClosed:
                if self._on_disconnect:
                    async_callback = threading.Thread(target = self._on_disconnect, daemon = True)
                    async_callback.start()

                raise NoConnectionError('Unexpectedly losed the connection while orchestrating queues and exchange for delegation')

    def broadcast(self, route, message, options = None):
        """ Broadcast a message to a particular route.

            :param str route:    the route
            :param str message:  the message
            :param dict options: additional options for basic_publish
        """
        default_parameters = self._generate_default_publish_options(
            self._default_broadcasting_options,
            SHARED_TOPIC_EXCHANGE_NAME,
            route,
            message,
        )

        options = fill_in_the_blank(options or {}, default_parameters)

        if 'exchange' not in options or not options['exchange']:
            options['exchange'] = SHARED_TOPIC_EXCHANGE_NAME

        exchange_name = options['exchange']

        with active_connection(self.url, self._on_connect, self._on_disconnect) as channel:
            try:
                log('debug', 'Declaring a shared topic exchange')

                channel.exchange_declare(
                    exchange      = exchange_name,
                    exchange_type = 'topic',
                    passive       = False,
                    durable       = True,
                    auto_delete   = False,
                )

                log('debug', 'Declared a shared topic exchange')

                log('debug', 'Broadcasting: route={} message={} options={}'.format(route, message, options))
                channel.basic_publish(**options)
                log('debug', 'Broadcasted: route={} message={} options={}'.format(route, message, options))

            except ConnectionClosed:
                if self._on_disconnect:
                    async_callback = threading.Thread(target = self._on_disconnect, daemon = True)
                    async_callback.start()

                raise NoConnectionError('Unexpectedly losed the connection while broadcasting an event')

    def observe(self, route, callback, resumable, distributed, options = None,
                simple_handling = True, controller_id = None):
        consumer_class = Consumer

        for overriding_consumer_class in self._consumer_classes:
            if overriding_consumer_class.can_handle_route(route):
                consumer_class = overriding_consumer_class

                break

        if not controller_id:
            controller_id = str(uuid.uuid4())

            log('info', 'Observer on {} will have the self-assigned controller ID {}'.format(route, controller_id))

        default_options = self._default_consuming_distributed_queue_options if distributed else self._default_consuming_shared_queue_options
        given_options   = options or {}

        queue_options    = fill_in_the_blank(given_options.get('queue',    {}), default_options.get('queue',    {}))
        exchange_options = fill_in_the_blank(given_options.get('exchange', {}), default_options.get('exchange', {}))

        parameters = dict(
            url               = self.url,
            route             = route,
            callback          = callback,
            shared_stream     = self._shared_stream,
            resumable         = resumable,
            distributed       = distributed,
            queue_options     = queue_options,
            simple_handling   = simple_handling,
            unlimited_retries = self._unlimited_retries,
            on_connect        = self._on_connect,
            on_disconnect     = self._on_disconnect,
            on_error          = self._on_error,
            controller_id     = controller_id,
            exchange_options  = exchange_options,
            auto_acknowledge  = self._auto_acknowledge,
            send_sigterm_on_disconnect = self._send_sigterm_on_disconnect,
        )

        consumer = consumer_class(**parameters)

        self._consumers.append(consumer)

        consumer.start()

        return consumer

    def _generate_default_publish_options(self, default_publishing_options, default_exchange_name,
                                          route, message):
        return {
            'exchange'    : default_publishing_options.get('exchange', default_exchange_name),
            'routing_key' : route,
            'body'        : json.dumps(message),
            'properties'  : BasicProperties(content_type = 'application/json'),
        }



import json
import logging
import pprint

from gallium.interface import ICommand

from .drivers.rabbitmq import Driver
from .observer         import Core, Observer, SYNC_START
from .helper           import prepare_logger


class SampleObserveWithDefaultOptions(ICommand):
    """ Run the sample observer with default options """
    def identifier(self):
        return 'sample.observe.default'

    def define(self, parser):
        parser.add_argument(
            '--debug',
            '-d',
            action = 'store_true'
        )

        parser.add_argument(
            '--bind-url',
            '-b',
            default='amqp://guest:guest@127.0.0.1:5672/%2F'
        )

    def execute(self, args):
        logger = prepare_logger(logging.DEBUG if args.debug else logging.INFO)

        def handle_driver_event(level, label, c):
            if not c or not hasattr(c, 'route') or not hasattr(c, 'queue_name'):
                getattr(logger, level.lower())(label)

                return

            getattr(logger, level.lower())('{} ({}): {}'.format(
                c.route,
                c.queue_name,
                label,
            ))

        driver = Driver(
            args.bind_url,
            unlimited_retries = True,
            on_connect        = lambda c = None:           handle_driver_event('info',  '--><-- CONNECTED',    c),
            on_disconnect     = lambda c = None:           handle_driver_event('error', '-x--x- DISCONNECTED', c),
            on_error          = lambda c = None, e = None: handle_driver_event('error', '-->-x- ERROR',        c),
        )

        service = Observer(driver)

        # In this example, delegation is disabled.
        # vireo.open('vireo.sample.primary', delegation_ttl = 5000)
        # vireo.on('vireo.sample.primary.delegated', lambda x: print('vireo.sample.primary.delegated: {}'.format(x)))

        def wrapper(label, data):
            print('[SAMPLE] {}:'.format(label))

            pprint.pprint(data, indent = 2)

        service.on('vireo.sample.direct',           lambda x: wrapper('vireo.sample.direct',    x))
        service.on('vireo.sample.secondary',        lambda x: wrapper('vireo.sample.secondary', x))
        service.on('vireo.sample.direct.resumable', lambda x: wrapper('vireo.sample.direct',    x), resumable = True)

        # With custom TOPIC exchange
        service.on(
            'vireo.sample.custom_topic_exchange.route_1',
            lambda x: wrapper('vireo.sample.custom_topic_exchange', x),
            options = {
                'exchange': {
                    'name': 'vireo_sample_topic_exchange',
                    'type': 'topic',
                }
            }
        )

        service.on(
            'vireo.sample.custom_topic_exchange.route_2',
            lambda x: wrapper('vireo.sample.custom_topic_exchange', x),
            options = {
                'exchange': {
                    'name': 'vireo_sample_topic_exchange',
                    'type': 'topic',
                }
            }
        )

        # With custom FANOUT exchange
        service.on(
            'vireo.sample.custom_fanout_exchange_1',
            lambda x: wrapper('vireo.sample.custom_fanout_exchange_1', x),
            options = {
                'exchange': {
                    'name': 'vireo_sample_fanout_exchange',
                    'type': 'fanout',
                }
            }
        )

        service.on(
            'vireo.sample.custom_fanout_exchange_2',
            lambda x: wrapper('vireo.sample.custom_fanout_exchange_2', x),
            options = {
                'exchange': {
                    'name': 'vireo_sample_fanout_exchange',
                    'type': 'fanout',
                }
            }
        )

        # With handling errors
        def error_demo(x):
            if 'e' in x:
                raise RuntimeError('Intentional Error')

            wrapper('vireo.sample.error', x)

        service.on('vireo.sample.error', error_demo)


        service.on_broadcast('vireo.sample.broadcast.one', lambda x: wrapper('vireo.sample.broadcast.one', x))
        service.on_broadcast('vireo.sample.broadcast.two', lambda x: wrapper('vireo.sample.broadcast.two', x))

        service.join(SYNC_START)


class SampleObserveWithCustomOptions(ICommand):
    """ Run the sample observer with custom options """
    def identifier(self):
        return 'sample.observe.custom'

    def define(self, parser):
        parser.add_argument(
            '--debug',
            '-d',
            action = 'store_true'
        )

        parser.add_argument(
            '--bind-url',
            '-b',
            default='amqp://guest:guest@127.0.0.1:5672/%2F'
        )

    def execute(self, args):
        logger = prepare_logger(logging.DEBUG if args.debug else logging.INFO)

        def handle_driver_event(level, label, c):
            if not c or not hasattr(c, 'route') or not hasattr(c, 'queue_name'):
                getattr(logger, level.lower())(label)

                return

            getattr(logger, level.lower())('{} ({}): {}'.format(
                c.route,
                c.queue_name,
                label,
            ))

        driver = Driver(
            args.bind_url,
            unlimited_retries = True,
            on_connect        = lambda c = None:           handle_driver_event('info',  '--><-- CONNECTED',    c),
            on_disconnect     = lambda c = None:           handle_driver_event('error', '-x--x- DISCONNECTED', c),
            on_error          = lambda c = None, e = None: handle_driver_event('error', '-->-x- ERROR',        c),
            default_consuming_shared_queue_options = {
                'exchange': {
                    'name': 'vireo_sample_custom_default_exchange',
                    'type': 'topic',
                },
            },
        )

        service = Observer(driver)

        def wrapper(label, data):
            print('[SAMPLE] {}:'.format(label))

            pprint.pprint(data, indent = 2)

        # With custom TOPIC exchange
        service.on(
            'vireo.sample_custom.route_1',
            lambda x: wrapper('vireo.sample_custom.route_1', x),
        )

        service.on(
            'vireo.sample_custom.route_2',
            lambda x: wrapper('vireo.sample_custom.route_2', x),
        )

        service.join(SYNC_START)


class SamplePublishWithCustomOptions(ICommand):
    """ Run the sample observer with custom options """
    def identifier(self):
        return 'sample.publish.custom'

    def define(self, parser):
        parser.add_argument(
            '--debug',
            '-d',
            action = 'store_true'
        )

        parser.add_argument(
            '--bind-url',
            '-b',
            default='amqp://guest:guest@127.0.0.1:5672/%2F'
        )

        parser.add_argument(
            'event_name',
            help = 'The name of the event (e.g., "sample.primary")'
        )

        parser.add_argument(
            'event_data',
            help  = 'The JSON-compatible string data of the event',
            nargs = '?'
        )

    def execute(self, args):
        prepare_logger(logging.DEBUG if args.debug else logging.INFO)

        driver = Driver(
            args.bind_url,
            default_publishing_options = {
                'exchange': 'vireo_sample_custom_default_exchange',
            },
        )

        service = Core(driver)

        service.emit(args.event_name, json.loads(args.event_data) if args.event_data else None)



from uuid import uuid4

from .core   import Core
from .helper import log
from .model  import RemoteSignal

ASYNC_START = 1
SYNC_START  = 2


class UnknownRunningModeError(RuntimeError):
    """ Error for unknown running mode """


class Observer(Core):
    """ Event Observer """
    def __init__(self, driver):
        Core.__init__(self, driver)

        self._identifier = str(uuid4())

        self._broadcast_event_to_observer_map = {}
        self._normal_event_to_observer_map    = {}

        log('debug', 'Observer ID: {}'.format(self.id))

    @property
    def id(self) -> str:
        """ Observer Identifier """
        return self._identifier

    def on(self, event_name, callback, resumable = False, simple_handling = True, options = None):
        """ Listen to an event with a callback function.

            :param str event_name: the name of the event
            :param callable callback: the callback callable
            :param bool resumable: the flag to indicate whether the event consumption can be resumed (as the data stream will never be deleted).
            :param bool simple_handling: the flag to instruct the code to return the content of the message, instead of returning the whole :class:`vireo.model.Message` object.

            The callback is a callable object, e.g., function, class method, lambda object, which
            takes only one parameter which is a JSON-decoded object.

            For example,

            .. code-block:: python

                def on_foo(self, message):
                    print('on_foo:', message)

                app.on('foo', on_foo)
                app.on('foo.lambda', lambda x: print('foo_lambda:', x))

            Here is an example for ``error_handler``.

            .. code-block:: Python

                def error_handler(consumer, exception):
                    ...
        """
        internal_observer = self._driver.observe(event_name, callback, resumable, False, options,
                                                 simple_handling = simple_handling,
                                                 controller_id = self.id)

        self._register_event_handler(self._normal_event_to_observer_map, event_name, internal_observer)

        return internal_observer

    def on_broadcast(self, event_name, callback, simple_handling = True, options = None):
        """ Listen to an distributed event with a callback function.

            :param str event_name: the name of the event
            :param callable callback: the callback callable
            :param bool simple_handling: the flag to instruct the code to return the content of the message, instead of returning the whole :class:`vireo.model.Message` object.

            The callback is a callable object, e.g., function, class method, lambda object, which
            takes only one parameter which is a JSON-decoded object.

            For example,

            .. code-block:: python

                def on_foo(self, message):
                    print('on_foo:', message)

                app.on('foo', on_foo)
                app.on('foo.lambda', lambda x: print('foo_lambda:', x))

            Here is an example for ``error_handler``.

            .. code-block:: Python

                def error_handler(consumer, exception):
                    ...
        """
        internal_observer = self._driver.observe(event_name, callback, False, True, options,
                                                 simple_handling = simple_handling,
                                                 controller_id = self.id)

        self._register_event_handler(self._broadcast_event_to_observer_map, event_name, internal_observer)

        return internal_observer

    def pause_on(self, event_name, remote_identifier = None):
        if not remote_identifier:
            self._remote_to_all_observers_on(self._normal_event_to_observer_map, event_name, 'pause')

            return

        self.emit(
            event_name,
            {
                'remote_signal' : RemoteSignal.PAUSE,
                'controller_id' : remote_identifier,
            }
        )

    def resume_on(self, event_name, remote_identifier = None):
        if not remote_identifier:
            self._remote_to_all_observers_on(self._normal_event_to_observer_map, event_name, 'resume')

            return

        self.emit(
            event_name,
            {
                'remote_signal' : RemoteSignal.RESUME,
                'controller_id' : remote_identifier,
            }
        )

    def pause_on_broadcast(self, event_name, remote_identifier = None):
        if not remote_identifier:
            self._remote_to_all_observers_on(self._broadcast_event_to_observer_map, event_name, 'pause')

            return

        self.broadcast(
            event_name,
            {
                'remote_signal' : RemoteSignal.PAUSE,
                'controller_id' : remote_identifier,
            }
        )

    def resume_on_broadcast(self, event_name, remote_identifier = None):
        if not remote_identifier:
            self._remote_to_all_observers_on(self._broadcast_event_to_observer_map, event_name, 'resume')

            return

        self.broadcast(
            event_name,
            {
                'remote_signal' : RemoteSignal.RESUME,
                'controller_id' : remote_identifier,
            }
        )

    def join(self, running_mode = SYNC_START):
        """ Wait for all handlers to stop.

            There are two mode: synchronous (``vireo.observer.SYNC_START``) and asynchronous
            (``vireo.observer.ASYNC_START``) joins.

            .. code-block:: python

                app.join(ASYNC_START)
        """
        if running_mode == ASYNC_START:
            self._driver.setup_async_cleanup()

            return

        if running_mode == SYNC_START:
            self._driver.join()

            return

        raise UnknownRunningModeError(running_mode)

    def stop(self):
        """ Send the signal to all handlers to stop observation.

            .. warning:: This method does not block the caller thread while waiting all handlers to stop.

            .. code-block:: python

                app.stop()
        """
        if self._driver:
            self._driver.stop_consuming()

    def _register_event_handler(self, observer_map, event_name, observer):
        if event_name not in observer_map:
            observer_map[event_name] = []

        observer_map[event_name].append(observer)

    def _find_all_event_handlers(self, observer_map, event_name, observer):
        if event_name not in observer_map:
            return []

        return observer_map[event_name]

    def _remote_to_all_observers_on(self, observer_map, event_name, remote_command):
        internal_observers = self._find_all_event_handlers(observer_map, event_name)

        for internal_observer in internal_observers:
            if not hasattr(internal_observer, remote_command):
                continue

            actual_remote_command = getattr(internal_observer, remote_command)

            if not callable(actual_remote_command):
                continue

            actual_remote_command()



from distutils.core import setup

setup(
    name         = 'vireo',
    version      = '0.11.2',
    description  = 'A library and framework for event-driven application development',
    license      = 'MIT',
    author       = 'Juti Noppornpitak',
    author_email = 'juti_n@yahoo.co.jp',
    # url          = 'http://shiroyuki.com/projects/vireo.html',
    packages     = [
        'vireo',
        'vireo.drivers',
        'vireo.drivers.rabbitmq',
    ],
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries'
    ],
    install_requires = ['pika']
)
